/*
Broilerplate GLUT/OpenGL main routine.

Orion Sky Lawlor, olawlor@acm.org, 2003/10/14
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ogl/ext.h"
#include "ogl/program.h"
#include "ogl/util.h"
#include "ogl/readback.h"
#include "ogl/main.h"
#include "osl/random.h"
#include "osl/random_vector.h"

#include "osl/random.cpp" /* For easy linking */
#include "ogl/easylink.h"

#include <vector>

oglTexture *decal; /* shadow decal */
double groundZ=0.0;
class myDrawable : public oglDrawable {
	oglReadback rb;
	
	// Information about a single sphere
	class sphereInfo {
	public:
		double radius;
		osl::Vector3d center;
		osl::Vector3d velocity;
		osl::Vector3d color;
		osl::Axes3d orientation;
		osl::Vector3d angVel;
		sphereInfo(osl::Random &r)
			:radius(0.03+r.nextFloat()*0.05),
			center(osl::nextCubeVector3d(r)+osl::Vector3d(0,0,1+radius)),
			velocity(osl::nextDirection(r)),
			color(osl::nextCubeVector3d(r)),
			angVel(osl::nextCubeVector3d(r))
		{
			orientation.nudge((r.nextFloat()-0.5)*2,(r.nextFloat()-0.5)*2);
			orientation.rotate((r.nextFloat()-0.5)*2);
		}
		void draw(void);
		void drawDecal(void); ///< Decal shadow
		void drawSilhouette(bool endCaps);
		void animate(double dt);
	};
	std::vector<sphereInfo> spheres;
	
	void drawScene(void);
	void drawObjects(void);
	void drawGround(void);
	
	void drawProjected(void);
	void drawProjectedStencil(void);
	void drawSilhouettes(GLenum drawSide,bool endCaps=false);
	void drawShadowMaps(const osl::Viewpoint &v);
	void drawShadowVolumes(void);
	void initSpheres(void);
public:
	myDrawable();
	void draw(const osl::Viewpoint &v);
}; 


void myDrawable::initSpheres(void) {
	spheres.erase(spheres.begin(),spheres.end());
	osl::Random48 r;
	for (unsigned int i=0;i<100;i++) 
		spheres.push_back(sphereInfo(r));
}

myDrawable::myDrawable() 
{
	initSpheres();
	
	/* Make shadow decal: a central dark spot
	  with fuzzy edges.
	*/
	const int w=32, h=32;
	unsigned char decalArr[w*h];
	for (int y=0;y<h;y++)
	for (int x=0;x<w;x++) {
		double dx=x/(double)w-0.5, dy=y/(double)h-0.5;
		double r=sqrt(dx*dx+dy*dy);
		double out=1.0;
		double r_min=0.2, r_max=0.4;
		if (r<r_min) out=0.0;
		else if (r>r_max) out=1.0;
		else out=(r-r_min)/(r_max-r_min);
		decalArr[y*w+x]=(unsigned char)(255*out);
	}
	decal=new oglTexture(decalArr,w,h,
		oglTexture_mipmap, GL_LUMINANCE,GL_UNSIGNED_BYTE,
		GL_LUMINANCE8);
}

void myDrawable::draw(const osl::Viewpoint &v)
{
	if (oglToggles['T'] || oglToggles['F']) /* Teapots-- only draw a few objects */
		spheres.erase(spheres.begin()+10,spheres.end());
	
	glDisable(GL_CULL_FACE);
	GLfloat zeros[4]={0.0f,0.0f,0.0f,0.0f};
	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, zeros);
	
/* Draw the main geometry */
	if (oglToggles['M']) /* shadow maps */
		drawShadowMaps(v);
	else if (oglToggles['V']) /* shadow volumes */
		drawShadowVolumes();
	else if (oglToggles['P']) { /* projective shadows *without* stencil */
		drawScene();
		drawProjected();
	}
	else if (oglToggles['L']) 
	{ /* projective shadows *with* stencil */
		drawScene();
		drawProjectedStencil();
	}
	else /* Don't draw shadows-- draw normally */ {
		drawScene();
		if (!oglToggles['d']) { /* Decal shadows */
			for (unsigned int i=0;i<spheres.size();i++) 
				spheres[i].drawDecal();
		}
	}

/* animate spheres */
	if (oglToggles['a']) { 
		static double lastAnim=oglTimer();
		double del=oglTimer()-lastAnim;
		if (del>1.0) del=1.0e-3;
		lastAnim=oglTimer();
		del*=1.0; /* set time scale */
		if (oglToggles['A']) del*=0.01; /* Bullet time */
		for (unsigned int i=0;i<spheres.size();i++) 
			spheres[i].animate(del);
	}
	if (oglToggles['R']) { /* reset spheres */
		initSpheres();
		oglToggles['R']=0;
	}
	
	oglRepaint(10);
}
/* Draw all the geometry in the scene */
void myDrawable::drawScene(void)
{
	drawGround();
	drawObjects();
}

void myDrawable::drawGround(void) {
	glColor3f(0.6,0.6,0.6);
	glNormal3f(0,0,1);
	glDisable(GL_TEXTURE_2D);
	glBegin(GL_QUAD_STRIP);
	double s=20.0; // ground size
	glVertex3f(-s,-s,0);
	glVertex3f(1+s,-s,0);
	glVertex3f(-s,1+s,0);
	glVertex3f(1+s,1+s,0);
	glEnd();
}
void myDrawable::drawObjects(void) {
	for (unsigned int i=0;i<spheres.size();i++) 
		spheres[i].draw();
}

void myDrawable::sphereInfo::draw(void)
{
	glPushMatrix();
	glTranslated(center.x,center.y,center.z);
	glColor3f(color.x,color.y,color.z);
	glDisable(GL_TEXTURE_2D);
	if (!oglToggles['T']) { /* Sphere */
		glutSolidSphere(radius,20,10);
	} else { /* Teapot */
		double m[16]={0.0};
		for (int axis=0;axis<3;axis++) {
			m[4*axis+0]=orientation.getX()[axis]; 
			m[4*axis+1]=orientation.getY()[axis]; 
			m[4*axis+2]=orientation.getZ()[axis]; 
			m[4*axis+3]=0.0;
		}
		m[15]=1.0;
		glMultMatrixd(m);
		glutSolidTeapot(5*radius);
	}
	glPopMatrix();
}

void myDrawable::sphereInfo::animate(double dt) 
{
	orientation.nudge(dt*angVel.x,dt*angVel.y); // HACK
	
	velocity.z-=dt*9.8; /* Gravitational accelleration pulls velocity downward */
	center+=dt*velocity;
	if (center.z<groundZ+radius && velocity.z<0) {
		center.z=groundZ+radius;
		velocity.z*=-0.93;
		velocity*=0.98; /* lose energy at each hit */
		angVel.y=-angVel.y; /* Spin along a different axis */
		angVel*=0.4;
	}
}


/** Save all matrices and reset to identity */
void oglMatrixSave(void) {
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
}
/** Restore all matrices and reset to identity */
void oglMatrixRestore(void) {
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	
	glPopAttrib();
}

/// Draw a screen-filling quad.
void oglDrawScreenQuad(void) {
	oglMatrixSave(); 
	glDisable(GL_LIGHTING);
	glBegin(GL_QUAD_STRIP); 
	glVertex3i(+1,-1,-1);  glVertex3i(+1,+1,-1);
	glVertex3i(-1,-1,-1);  glVertex3i(-1,+1,-1);
	glEnd();
	glEnable(GL_LIGHTING);
	oglMatrixRestore();
}

/***************** Decal Shadows ('d', 'P' and 'L') *****************/

// Draw a decal shadow underneath each object.
void myDrawable::sphereInfo::drawDecal(void) {
	osl::Vector3d v=center;
	v.z=groundZ+0.002; /* project down to floor (almost!) */
	double s=1.7*radius; /* size of shadow */
	
	glDisable(GL_LIGHTING);
	glColor4f(1,1,1,1);
	glBlendFunc(GL_DST_COLOR, GL_ZERO); /* out=src*dest */
	glDepthMask(GL_FALSE); /* don't mess up depth buffer */
	oglTextureQuad(*decal,
		  v+s*osl::Vector3d(-1,+1,0),v+s*osl::Vector3d(+1,+1,0),
		  v+s*osl::Vector3d(-1,-1,0),v+s*osl::Vector3d(+1,-1,0));
	glDepthMask(GL_TRUE);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); /* out=lerp(src,dest,alpha) */
	glEnable(GL_LIGHTING);
}

// Draw world's geometry, projected to the ground plane.
void myDrawable::drawProjected(void) {
	/* Make a matrix to squish Z down to the ground plane */
	glPushMatrix();
	double m[16]={0.0};
	m[4*0+0]=m[4*1+1]=m[4*3+3]=1.0; /* Don't mess with X,Y, or W */
	m[4*2+2]=0; m[4*3+2]=groundZ+0.002; /* Squish Z */
	glMultMatrixd(m);
	
	/* World's stupidest way to override sphere colors: a fragment program */
	static oglProgram fp(
		"!!ARBfp1.0\n"
		"MOV result.color,{0,0,0,0.9};\n"
		"END"
	);
	glDepthMask(GL_FALSE); /* don't mess up depth buffer */
	fp.begin();
	drawObjects();
	fp.end();
	glDepthMask(GL_TRUE);
	
	glPopMatrix();
}

// As with drawProjected, but only draws shadows once, using stencil buffer.
void myDrawable::drawProjectedStencil(void)
{
	glEnable(GL_STENCIL_TEST); /* needed for stencil buffer *writes* to work! */
	
/* Set the stencil for all the projected geometry */
	glStencilFunc(GL_ALWAYS,1,0xff);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
	glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE); /* don't mess with color */
	drawProjected();
	
// Darken the shadows, by drawing where the stencil is >0
	glStencilFunc(GL_NOTEQUAL,0,0xff); 
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE); /* allow changes to color */
	glColor4f(0,0,0,0.9);
	oglDrawScreenQuad();
	
	glDisable(GL_STENCIL_TEST);
}

/********************* Shadow Maps ('M') **********************/

void myDrawable::drawShadowMaps(const osl::Viewpoint &v) {

/* Draw world from light's point of view into shadow map */
	oglMatrixSave();
	
	/* Map world XYZ to shadow map clip coordinates.
	  World XYZ of interest ranges from iMin to iMin+iDel on each axis;
	  Clip coordinates range from [-1,+1] on each axis.
	*/
	int shadow_wid=512, shadow_ht=512;
	if (oglToggles['r']) shadow_wid=shadow_ht=32; /* lo-res version */
	double xMin=-1.5, xDel=4.0;
	double yMin=xMin, yDel=xDel;
	double zMin=groundZ-0.1, zDel=3.0;
	double xScale=2.0/xDel, yScale=2.0/yDel, zScale=2.0/zDel;
	osl::ViewMatrix3d sm;
	double w_scale=1.0e20;
	sm.setRow(0,w_scale*osl::Vector3d(xScale,0,0),w_scale*(-1-xMin*xScale));
	sm.setRow(1,w_scale*osl::Vector3d(0,yScale,0),w_scale*(-1-yMin*yScale));
	sm.setRow(2,w_scale*osl::Vector3d(0,0,zScale),w_scale*(-1-zMin*zScale));
	sm.setRow(3,osl::Vector3d(0,0,0),w_scale*1);
	double shadowMapMatrix[16]; sm.makeOpenGL(shadowMapMatrix);
	glLoadMatrixd(shadowMapMatrix);
	glViewport(0,0,shadow_wid,shadow_ht);
	
	// Draw the world, to generate depth image
	glDepthFunc(GL_GEQUAL);/* because bigger Z means higher now */
	glClearDepth(0.0f); /* Depth should now start at zero (not 1) */
	glClear(GL_DEPTH_BUFFER_BIT);
	
	drawScene();
	
// Read back depth from light source and upload to a texture
	rb.readDepth(0,0,shadow_wid,shadow_ht);
	oglTexture *depthTex=rb.getTex();
	
	glDepthFunc(GL_LEQUAL);/* back to normal */
	glClearDepth(1.0f);
	
	oglMatrixRestore();
	
	if (oglToggles['s']) return; /* just show shadow map image */
	
// Clear the screen (otherwise light source junk stays there)
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	
	if (oglToggles['t']) { /* Just show depth buffer texture */
		glColor4f(1,1,1,1);
		zMin=groundZ;
		oglTextureQuad(*depthTex,
			osl::Vector3d(xMin,yMin+yDel,zMin), osl::Vector3d(xMin+xDel,yMin+yDel,zMin),
			osl::Vector3d(xMin,yMin,zMin), osl::Vector3d(xMin+xDel,yMin,zMin));
		return;
	}
	
/* Finally, turn on fragment program and render world
  using the shadow map. */
	static oglProgram vp(oglProgram_vertexStandard "END",GL_VERTEX_PROGRAM_ARB);
	static oglProgram fp(oglProgram_fragmentStandard
	"PARAM scales=program.env[0];\n"
	"PARAM offsets=program.env[1];\n"
	"TEMP lc,lighting,diffuse;\n"
	"MAD lc, world,scales,offsets;\n" // Shadow coordinates
	"TEX lighting, lc, texture[0], 2D;\n" // Shadow map lookup-- find lit depth
	"ADD lc.z, lc.z, 0.001;\n" // Avoid roundoff in shadow lookup 
	"SGE lighting, lc.z, lighting;\n" // If you're >= shadow map, you're lit!
	"DP3_SAT diffuse, worldN, state.light[0].position;\n"
	"MUL lighting,lighting,diffuse;\n" // Modulate lighting by normal factor
	"MAD lighting,lighting,0.8,0.1;\n" // Scale down diffuse and add ambient
	"MOV lighting.a,1.0;\n"
	"MUL result.color, lighting, albedo;\n"
	"END"
	);
	
	depthTex->bind();
	glProgramEnvParameter4dvARB(GL_VERTEX_PROGRAM_ARB, 0, v.getEye());
	osl::Vector3d scales(1.0/xDel,1.0/yDel,1.0/zDel);
	osl::Vector3d offsets(-xMin*scales.x,-yMin*scales.y,-zMin*scales.z);
	glProgramEnvParameter4dvARB(GL_FRAGMENT_PROGRAM_ARB, 0, scales);
	glProgramEnvParameter4dvARB(GL_FRAGMENT_PROGRAM_ARB, 1, offsets);
	vp.begin(); fp.begin();
	drawScene();
	vp.end(); fp.end();
}


/****************** Shadow Volumes ('V') ******************/
void myDrawable::drawShadowVolumes(void) {
// Draw geometry normally
	drawScene();

// Set up stencil buffer with location of shadows
	glDisable(GL_LIGHTING); /* don't do lighting */
	glDepthMask(GL_FALSE); /* don't mess up z buffer */
	glEnable(GL_STENCIL_TEST); /* needed for stencil buffer *writes* to work! */
	glStencilFunc(GL_ALWAYS,1,0xff);
	
	/* Render shadow silhouettes into stencil buffer */
	if (oglToggles['s']) {
		/* Draw silhouettes onscreen, in yellow */
		glColor4f(1,1,0,0.2);
	} else 
	{ /* Regular shadow volumes */
		glColorMask(GL_FALSE,GL_FALSE,GL_FALSE,GL_FALSE); /* don't mess with color */
	}
	
	if (!oglToggles['c']) 
	{ /* Normal shadow volumes: */
		/* Increment all the front faces */
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR); 
		drawSilhouettes(GL_FRONT);
		/* Decrement all the back faces */
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
		drawSilhouettes(GL_BACK);
	} 
	else
	{ /* Use Carmack reversing trick */
		glDepthFunc(GL_GREATER); /* flip the sense of the depth compare */
		/* Increment all the back faces that are hidden */
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR); 
		drawSilhouettes(GL_BACK,true);
		/* Decrement all the front faces that are hidden */
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
		drawSilhouettes(GL_FRONT,true);
		glDepthFunc(GL_LESS);
	} 
	
	glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);

// Darken the shadows, by drawing where the stencil is >0
//  FIXME: chould instead first draw ambient, then add diffuse where stencil is *0*.
	if (!oglToggles['s']) {
		glStencilFunc(GL_NOTEQUAL,0,0xff); /* draw where stencil is not zero */
		glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP); 
		glColor4f(0,0,0,0.9);
		oglDrawScreenQuad();
	}
	
	glDisable(GL_STENCIL_TEST);
	glEnable(GL_LIGHTING);
	glDepthMask(GL_TRUE);
}

void myDrawable::drawSilhouettes(GLenum drawSide,bool endCaps) 
{
	/* Expensive hack: draws all faces, but culls out those facing wrong way! */
	glEnable(GL_CULL_FACE);
	if (oglToggles['b']) { /* don't draw back sides */
		if (drawSide==GL_BACK) return;
		if (drawSide==GL_FRONT_AND_BACK) drawSide=GL_FRONT;
	}
	switch (drawSide) {
	case GL_FRONT: glCullFace(GL_BACK); break; // to draw front, cull back
	case GL_BACK: glCullFace(GL_FRONT); break; // to draw back, cull front
	case GL_FRONT_AND_BACK: glDisable(GL_CULL_FACE); break;
	}
	for (unsigned int i=0;i<spheres.size();i++) 
		spheres[i].drawSilhouette(endCaps);
	glDisable(GL_CULL_FACE);
}

/** Extrude the silhouette of this sphere, a cylinder,
down to the ground plane (& beyond). 
*/
void myDrawable::sphereInfo::drawSilhouette(bool endCaps)
{
	glPushMatrix();
	double beyond=10; /* Depth beneath the ground to start silhouettes */
	double heightShift=groundZ-beyond;
	glTranslated(center.x,center.y,heightShift);
	static GLUquadric *q=0;
	if (q==0) q=gluNewQuadric();
	gluCylinder(q,radius,radius,center.z-heightShift,20,2);
	glPopMatrix();
	
	if (!oglToggles['e']) 
	/* Add end caps, so shadow silhouette shape is closed--
	  Avoids artifacts when looking from light source point of view.
	*/
	for (int end=0;end<2;end++) {
		glPushMatrix();
		if (end) { /* end cap on top */
			glTranslated(center.x,center.y,center.z);
			gluQuadricOrientation(q,GLU_OUTSIDE);
		} else { /* end cap on bottom */
			glTranslated(center.x,center.y,heightShift);
			gluQuadricOrientation(q,GLU_INSIDE);
		}
		gluDisk(q,0,radius,20,2);
		glPopMatrix();
	}
	gluQuadricOrientation(q,GLU_OUTSIDE);
}


/*************** Main **************/

int main(int argc,char **argv) {
	oglOptions i("OpenGL Demo");
	i.lighting=true;
	i.fullscreen=false;
	i.clearColor=osl::Vector3d(0.2,0.3,0.4);
	i.width=1000;
	i.height=700;
	i.depth=true;
	i.depthRange=100;
	i.stencil=true;
	oglSetup(i);
	oglSetupExt();
	
	oglEventloop(new oglTrackballController(120.0,4.0,osl::Vector3d(0.5,0.5,0.0))
		,new myDrawable);
	return 0;
}

